/*
 *  Tagging      Responses for Disaster Recovery
 *  Author:      Avinash Kulkarni
 *  Date:        05/09/2009
 *  Description: This program explores the possibility
 *               of tagging responses sent to clients
 *               which could later be used to recover
 *               from disastrous failures of primary.
 *               See 'ReadMe.txt' for details on this
 *               program.
 */

#include "async.h"
#include "cdb.h"
#include <map>
#include <iostream>


#define MAX_LISTEN_BACKLOG 5
#define MAX_CMD_LEN 128
#define MAX_DATA_LEN 8192
#define MAX_DATA (128+8192)

using namespace std;

static int clientSock;
static int clientListenSock;
static int local_port;
static int remote_port;

int fileOffset = 0;
int tLen = 100;

struct sockaddr_in remote_addr; /* The address of the target server */

timecb_t *delayTimer = 0;

char buf[MAX_CMD_LEN + MAX_DATA_LEN];
char *localFileName;
int fd;

map<int, char*> clientLog;
map<int, char*>::iterator mIter;

static void doClientWrites();
static void receivePrimaryRsp();

int lastAckedWrite = 0;

static void delayCB()
{    
    //warn << "In delay CB\n";
    fdcb(clientSock, selwrite, wrap(doClientWrites));    
}

static void waitForServer()
{
    fdcb(clientSock, selread, wrap(receivePrimaryRsp));
}

// Receive a response from the primary to a write that
// the client sent out. 
// Parse the contents of the received response and store
// each tagged message in a log
static void receivePrimaryRsp()
{
    int n = read(clientSock, buf, MAX_DATA);
    if ( n < 0 )
	{
	    // Close this connection off
	    shutdown(clientSock, SHUT_RDWR);
	    fdcb(clientSock, selread, 0);
	    close(clientSock);
	    warn << "Rcv Pri. Rsp: Read error\n";	    
	}
    else if ( n == 0 )
	{
	    warn << "Rcv. Pri. Rsp: Got no data??\n";
	    delaycb(60, 0, wrap(waitForServer));
	    fdcb(clientSock, selread, 0);
	}
    else
	{
	    //warn << "Read:" << n << " bytes from the primary\n";
	    char *temp = buf;
	    while (true )
	    {	
			cmdPacket *tmp = reinterpret_cast<cmdPacket*>(temp);	    
			if ( tmp->cmd == PRsp )
			{
				//warn << "Received a response from primary with LogIdx:"
				// << tmp->logIndex << "\n";

				if ( tmp->flushReq == true )
				{
					// Flush old log entries
					//warn << "Flushing old entries on Primary request\n";
					mIter = clientLog.begin();
					while ( mIter != clientLog.end())
					{
						delete mIter->second;
						mIter++;
					}
					clientLog.erase(clientLog.begin(), clientLog.end());
				}
				
				if ( clientLog.find(tmp->logIndex) == clientLog.end() )
				{
					// Save this last data entry into our log - only if previously not
					// received
					char *newBuf = new char[sizeof(cmdPacket)+100];
					memcpy(newBuf, temp, sizeof(cmdPacket)+100);
					clientLog[tmp->logIndex] = newBuf;
				}
				else
				{
					warn << "Received duplicate log entry\n";
				}
			}
			if ( tmp->moreToFollow == true )
			{
				temp += (sizeof(cmdPacket)+100);
			}
			else
			{
				break;
			}
		}
	    // Now, reenable writes to server - but wait 1 second
	    delaycb(1, 0, wrap(delayCB));
	    // Disable reads
	    fdcb(clientSock, selread, 0);		        
	}
}

// Write data out to the primary - 100 bytes at a time
static void doClientWrites()
{
    memset(buf, 0, sizeof(buf));
    cmdPacket *reg = new cmdPacket;
    memset(reg, 0, sizeof(cmdPacket));    
    reg->cmd = CWrite;
    char fileName[64];
    sprintf(fileName, "/tmp/fuse/%s", localFileName);
    strcpy(reg->fileName, fileName);
    reg->localPort = local_port;
    reg->offset = fileOffset;
    memcpy(buf, reg, sizeof(cmdPacket));
    delete reg;
    // Write the first 8192 bytes of a random text file
    char temp[100]; // Read 100 bytes from the file into this buffer
    memset(temp, 0, sizeof(temp));
    int n = read(fd, temp, 100 );
    if ( n < 0 )
    {
	// Retry later??
	warn << "Read from file failed!\n";
	fdcb(clientSock, selwrite, 0);
	delayTimer = delaycb(5, 0, wrap(delayCB));	
	return;
    }

    memcpy(buf+sizeof(cmdPacket), temp, n);
    // Now, write it out to the primary, and turn off
    // reads until another 5 seconds
    int numBytes = write(clientSock, buf, sizeof(cmdPacket) + n);
    if ( numBytes < 0 )
	{
	    close(fd);
	    shutdown(clientSock, SHUT_RDWR);
	    fdcb(clientSock, selread, 0);
	    close(clientSock);
	    fatal << "Write error:\n";		    
	}
	else if ( numBytes == 0 )
	{
	    warn << "Wrote 0 bytes out\n";
	    
	}
	else
	{
	    //warn << "Wrote:" << n << " bytes from offset:"
	    // << fileOffset << "\n";
	    lastAckedWrite = fileOffset;
	    fileOffset += n;
	    fdcb(clientSock, selwrite, 0);
	    // No delay needed here
	    //delayTimer = delaycb(10, 0, wrap(delayCB));
	    // Await response from the server
	    fdcb(clientSock, selread, wrap(receivePrimaryRsp));	    
	}    
}

    
// Receive a confirmation of registration request from the primary
static void recvRegnConf()
{
    memset(buf, 0, sizeof(buf));
    int n = read(clientSock, buf, sizeof(buf));
    if ( n < 0 )
    {
	shutdown(clientSock, SHUT_RDWR);
	fdcb(clientSock, selread, 0);
	close(clientSock);
	fatal << "Regn conf rcv: read error\n";	
    }
    else if ( n == 0 )
    {
	warn << "Regn conf rcv: read returned 0??\n";
    }
    else
    {
	//warn << "Reg. Recv conf: received " << n << " bytes from client \n";
	cmdPacket *cmd = reinterpret_cast<cmdPacket *>(buf);
	if ( cmd->cmd != RegisterDone )
	{
	    // Close this connection off
	    shutdown(clientSock, SHUT_RDWR);
	    fdcb(clientSock, selread, 0);
	    close(clientSock);	    
	    fatal << "Regn. Recv conf: invalid packet received\n";
	}
	else
	{
	    //warn << "Regn Recv. conf. recvd.!\n";
	    // Enable writes on this socket so we could send data to primary
	    fdcb(clientSock, selread, 0);
	    fdcb(clientSock, selwrite, wrap(doClientWrites));
	    fd = open(localFileName, O_RDONLY);
	    if ( fd <= 0 )
	    {
		// Close this connection off
		shutdown(clientSock, SHUT_RDWR);
		fdcb(clientSock, selread, 0);
		close(clientSock);	    		
		fatal << "Read from file failed!\n";
	    }
	    
	    // Also, enable listen on the local port
	    listen(clientSock, MAX_LISTEN_BACKLOG);
	}
    }
}


// Now that the client has connected to the primary,
// send a registration request to the primary
static void registerWithServer()
{
    // Send registration information
    memset(buf, 0, sizeof(cmdPacket));
    cmdPacket *reg = new cmdPacket;
    memset(reg, 0, sizeof(cmdPacket));    
    reg->cmd = Register;
    reg->localPort = local_port + 1000;
    memcpy(buf, reg, sizeof(cmdPacket));
    int numBytes = write(clientSock, buf, sizeof(cmdPacket));
    if ( numBytes < 0 )
    {
	delete reg;
	fatal << "Register: Write returned 0: n:" << numBytes << "\n";
    }
    else if ( numBytes == 0 )
    {
	delete reg;
	warn << "Register: Write did not complete?? Retrying";
    }
    else
    {
	//warn << "Register: Write worked - sent " << numBytes
	//   << " to the primary\n";
	if ( (unsigned)numBytes < sizeof(cmdType))
	{
	    // Should have sent the whole packet - but we did not
	    // Close this socket for diag purpose
	    delete reg;
	    shutdown(clientSock, SHUT_RDWR);
	    fdcb(clientSock, selwrite, 0);
	    close(clientSock);
	    fatal << "Register: Closing on sent < sizeof\n";
	}
	else
	{
	    // Close the socket if equal size packet was sent
	    delete reg;
	    //warn << "Register: Sent register request\n";
	    // Nwo, unset write on this socket, and set read
	    // to await regn. conf
	    fdcb(clientSock, selwrite, 0);
	    fdcb(clientSock, selread, wrap(recvRegnConf));
	}
    }    
}

// Make a connection to the primary to initiate data transfers
static void connectToServer()
{    
    if ( connect(clientSock, (struct sockaddr *) &remote_addr, 
		 sizeof(struct sockaddr_in)) < 0)
    {
	if (errno != EINPROGRESS) /* nonblocking, connection stalled */
	{
	    fprintf(stderr, "connect error %s\n", strerror(errno));
	    fdcb(clientSock, selwrite, 0);
	    close(clientSock);
	    exit(1);
	}
    }
    else
    {
	//fprintf(stderr, "Connect succeeded\n");
	//fdcb(clientSock, selwrite, 0);
	fdcb(clientSock, selwrite, wrap(registerWithServer));
    }
}

// Send the contents of the client's log in response
// to a request from the primary
static void sendLogRsp(int primary_sock)
{
    // Each log entry is 228 bytes.
    // Generate a buffer of 228 * numLogEntries
    int numLogEntries = clientLog.size();
    if ( numLogEntries == 0 )
    {
	// Special case - just write a EmptyLog rsp to
	// the primary
	cmdPacket *tmp = new cmdPacket;
	tmp->cmd = EmptyLog;
	int n = write(primary_sock, tmp, sizeof(cmdPacket));
	if ( n < 0 )
	{
	    fatal << "Unable to write to primary\n";
	}
	else
	{
	    delete tmp;
	    warn << "Sent an empty log entry message to primary\n";
	    fdcb(primary_sock, selwrite, 0);
	    shutdown(primary_sock, SHUT_RDWR);
	    close(primary_sock);	    
	    return;
	}
    }
    char *rspBuf = new char[numLogEntries * (sizeof(cmdPacket) + 100)];
    memset(rspBuf, 0, (numLogEntries * (sizeof(cmdPacket) + 100)));

    mIter = clientLog.begin();
    char *bufPtr = rspBuf;
    while ( mIter != clientLog.end())
    {
	memcpy(bufPtr, mIter->second, (sizeof(cmdPacket) + 100));
	bufPtr += (sizeof(cmdPacket) + 100);
	mIter++;
    }
    // Now, write out this buffer to the primary
    int n = write(primary_sock, rspBuf, (numLogEntries * (sizeof(cmdPacket) + 100)));
    if ( n < 0 )
    {
	fatal << "Unable to write to primary\n";
    }
    else
    {
	//warn << "Sent:" << numLogEntries << " log entries to primary\n";
	delete rspBuf;
	fdcb(primary_sock, selwrite, 0);
	shutdown(primary_sock, SHUT_RDWR);
	close(primary_sock);	    
	return;
    }    
}


// Read data from socket that was created when primary
// initiated a connect operation during recovery
static void doPrimaryRead(int primary_sock)
{
    char buf[MAX_DATA];
    memset(buf, 0, sizeof(buf));
    int n = read(primary_sock, buf, MAX_DATA);
    if ( n < 0 )
    {
	fatal << "In Primary read - failed!\n";
    }
    else if ( n == 0 )
    {
	fatal << "In primary read - got 0 bytes??\n";
    }
    else
    {
	//warn << "In primary read - got " << n << " bytes from primary!\n";
	cmdPacket *tmp = reinterpret_cast<cmdPacket*>(buf);
	if ( tmp->cmd == LogQuery )
	{
	    // We need to build a log query response
	    // Disable reads on this socket now
	    fdcb(primary_sock, selread, 0);
	    // Enable write on this socket - we will generate
	    // the response packet in write
	    fdcb(primary_sock, selwrite, wrap(sendLogRsp, primary_sock));
	}
    }
}

// Accept connection from primary process during 
// a recovery session
static void primaryConn()
{
    warn << "Received connection from primary\n";
    int primary_sock;
    struct sockaddr_in proxy_addr, primary_addr;
    socklen_t addr_len = sizeof(struct sockaddr_in);
    memset(&proxy_addr, 0, sizeof(struct sockaddr_in));	
    if ((primary_sock = accept(clientListenSock, (struct sockaddr *) &proxy_addr, 
			      &addr_len)) < 0)
	{
	    if (errno != EAGAIN) /* nonblocking, retry */
		fprintf(stderr, "accept error %s\n", strerror(errno));
	}
    else
	{
	    make_async(primary_sock); 
	    addr_len = sizeof(struct sockaddr_in); /* in/out parameter */
	    if (getpeername(primary_sock, (struct sockaddr *) &primary_addr, 
			    &addr_len) < 0)
		{
		    fprintf(stderr, "getsockname error %s\n", strerror(errno));
		    exit(1);
		}
	    warn << "Received connection from:" << inet_ntoa(primary_addr.sin_addr)
		 << "\n";
	    // Enable read on this socket
	    fdcb(primary_sock, selread, wrap(doPrimaryRead, primary_sock));
	}
}

// Print the state of the log on the primary,
// and the last write from this client that was
// acknowledged by the primary
static void displayClientStatus()
{
    cout.flush();
    cout << "\33[H\33[2J";
    cout << "\t\t\tClient Status\n";
    cout << "\n";
    cout << "Client Log Content\n";
    cout << "==================\n";

    mIter = clientLog.begin();
    while ( mIter != clientLog.end())
    {
	cmdPacket *tmp = reinterpret_cast<cmdPacket*>(mIter->second);
	cout << "Log Entry Index:" << tmp->logIndex
	     << " Filename:" << tmp->fileName
	     << " Offset:" << tmp->offset
	     << "\n";
	mIter++;
    }

    cout << "\n";
    cout << "Last offset sent to primary:" << (fileOffset - 100)
	 << "\n";
    cout << "Last write ack'ed by primary:" << lastAckedWrite
	 << "\n";    
    delaycb(5, 0, wrap(displayClientStatus));    
    
}

// Start routine
// The client begins by connecting to the primary
// process on the socket specified in cmd. line param.
// After a successful connect, it issues writes from the
// file specified in cmd. line param and sends them out
// to the primary.
// The client also creates a socket that is set to be in
// listen mode. It awaits connections from primary during
// recovery mode.
int main(int argc, char **argv)
{

    struct sockaddr_in client_addr;
    char *remote_name;
    struct hostent *h;
    
    int arg_idx = 1;
    
    if (argc != 5)
    {
	fprintf(stderr, "Usage %s <remote-target> <remote-target-port> "
		"<local-port> <local-file-name>\n", argv[0]);
	exit(1);
    }

    cout << "\33[H\33[2J";
    cout.flush();    
    
    remote_name = argv[arg_idx++];
    remote_port = atoi(argv[arg_idx++]);    
    local_port = atoi(argv[arg_idx++]);
    localFileName = argv[arg_idx++];

    /* Lookup server name and establish control connection */
    if ((h = gethostbyname(remote_name)) == NULL)
    {
	fprintf(stderr, "gethostbyname(%s) failed %s\n", remote_name, 
		strerror(errno));
	exit(1);
    }
    memset(&remote_addr, 0, sizeof(struct sockaddr_in));
    remote_addr.sin_family = AF_INET;
    memcpy(&remote_addr.sin_addr.s_addr, h->h_addr_list[0], sizeof(in_addr_t));
    remote_addr.sin_port = htons(remote_port);        

    if ((clientSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0 )
	{
	    fatal << "Unable to create client socket: Exiting!\n";
	    exit(1);
	}

    memset(&client_addr, 0, sizeof(struct sockaddr_in));
    client_addr.sin_family = AF_INET;
    client_addr.sin_addr.s_addr = INADDR_ANY; /* bind to all local addresses */
    client_addr.sin_port = htons(local_port);
    if (bind(clientSock, (struct sockaddr *) &client_addr, 
	     sizeof(client_addr)) < 0)
	{
	    fprintf(stderr, "bind error %s\n", strerror(errno));
		exit(1);
	}    
    make_async(clientSock);
    /* register socket and signal handlers with the async event loop */
    fdcb(clientSock, selwrite, wrap(connectToServer));


    if ((clientListenSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0 )
	{
	    fatal << "Unable to create client socket: Exiting!\n";
	    exit(1);
	}

    // This will be the address this client will listen on
    // for requests from primary to regenerate data
    // Its port number is 1000 greater than the one it uses
    // to connect to the primary
    memset(&client_addr, 0, sizeof(struct sockaddr_in));
    client_addr.sin_family = AF_INET;
    client_addr.sin_addr.s_addr = INADDR_ANY; /* bind to all local addresses */
    client_addr.sin_port = htons(local_port + 1000);
    if (bind(clientListenSock, (struct sockaddr *) &client_addr, 
	     sizeof(client_addr)) < 0)
	{
	    fprintf(stderr, "bind error %s\n", strerror(errno));
		exit(1);
	}    
    make_async(clientListenSock);
    listen(clientListenSock, MAX_LISTEN_BACKLOG);
    fdcb(clientListenSock, selread, wrap(primaryConn));
    
    sigcb(SIGINT, wrap(exit, 1));
    sigcb(SIGTERM, wrap(exit, 1));

    delaycb(5, 0, wrap(displayClientStatus));
    
    amain(); /* async main */    
}
