/*
 *  Tagging      Responses for Disaster Recovery
 *  Author:      Avinash Kulkarni
 *  Date:        05/09/2009
 *  Description: This program explores the possibility
 *               of tagging responses sent to clients
 *               which could later be used to recover
 *               from disastrous failures of primary.
 *               See 'ReadMe.txt' for details on this
 *               program.
 */

#include "async.h"
#include "cdb.h"
#include <map>
#include <iostream>
#include <fstream>
#include <vector>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>


#define PRIMARY_PORT 1337 // The port that the clients can connect to primary on
#define MAX_LISTEN_BACKLOG 5
#define MAX_ADDR_NAME 64
#define MAX_CMD_LEN 128   // Max. length of the cmd. portion of a tagged response
#define MAX_DATA_LEN 8192 // Max. length of data portion of tagged response
#define MAX_DATA (MAX_DATA_LEN + MAX_CMD_LEN) // Max. length of tagged response


extern int errno;
using namespace std;

static int primarySock;
char buf[MAX_CMD_LEN + MAX_DATA_LEN];

// The port that the primary will use to connect
// to clients in order to obtain recovery data
int recovery_port = 1338;

// User configurable parameter that controls
// how many tags will be included in a response 
int numRspToClient = 0;

// Each client connected to the primary
// is represented by an instance of this class.
struct client
{
    int clientSock;
    char client_hname[MAX_ADDR_NAME];
    struct sockaddr_in client_addr;
    int clientPort;
    bool registered;
    bool flush;
    bool readDisabled;
    char buf[MAX_CMD_LEN + MAX_DATA_LEN];
};

// List of clients of the system
map<int, client*> clients;

// This will contain log entries that will be flushed 
// to the backup server at regular intervals
vector<char*> log;

// This counter is used to assign unique ids to each log 
// entry
int logCounter = 0;

// List of client sockets that the primary will connect to
// for recovery
vector<int> recoverySockList;

// Forward declaration
static void doClientReads(int);

// A map with an entry for each file that the
// primary is currently storing. The map stores
// the last offset that has been flushed to the backup
map<int, char*> flushOffsetTable;

map<char*, int> recoveredFiles;

// Flush writes from the log to the backup.
// At the end of this function call, primary
// and backup are sync'ed up with each other.
static void flushWrites()
{
	// Iterate through log contents, flushing each
	// log entry in turn
    //warn << "In flush writes\n";
    cmdPacket *tmp;    
    for ( unsigned int i = 0; i < log.size(); i++ )
	{
	    char *buf = log[i];
	    tmp = reinterpret_cast<cmdPacket*>(buf);
	    //warn << "Flushing Log entry to S3FS:" << " with log ID:"
	    // << tmp->logIndex << "\n";
	    //warn << "Filename:" << tmp->fileName << ", offset:" << tmp->offset
	    // << "\n";
	    int fd;
	    if ( tmp->offset == 0 )
			fd = open(tmp->fileName, O_CREAT|O_WRONLY|O_TRUNC, (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH | S_IWGRP | S_IWOTH));
	    else
			fd = open(tmp->fileName, O_WRONLY);
	    int s = 0;
	    if ( tmp->offset != 0 )
		s = lseek(fd, 0, SEEK_END);
	    write(fd, buf + sizeof(cmdPacket), 100);
	    char *str = new char[50];
	    sprintf(str, "%s-%d", tmp->fileName, tmp->offset);
	    flushOffsetTable[tmp->localPort] = str;	    
	    //warn << "Seek:" << s << ", Wrote:" << n << " bytes to " << tmp->fileName << "\n";
	    close(fd);
	    delete buf;
	}

    // Clear out the log
    log.erase(log.begin(), log.end());
    
    // Update flush table with new offsets
    map<int, client*>::iterator mIter;
    for ( mIter = clients.begin(); mIter != clients.end(); mIter++)
	{
	    client *c = mIter->second;
	    c->flush = true;
	    fdcb(mIter->first, selread, wrap(doClientReads, mIter->first));
	}
	// Schedule another callback in 120 seconds
    delaycb(120, 0, wrap(flushWrites));
}

// Display the status of primary on the screen
static void displayPrimaryStatus()
{
   map<int, client*>::iterator mIter;
   for ( mIter = clients.begin(); mIter != clients.end(); mIter++)
       {
	   client *c = mIter->second;
	   if ( c->readDisabled )
	       {
		   fdcb(mIter->first, selread, wrap(doClientReads, mIter->first));
		   c->readDisabled = false;
	       }
       }

   // Print the clients connected to the primary
   cout << "\33[H\33[2J";
   cout << "\t\tCurrent Status of the Primary\n";
   cout << "Current Clients:\n";
   cout << "================\n";
   for ( mIter = clients.begin(); mIter != clients.end(); mIter++)
       {
	   client *c = mIter->second;
	   cout << inet_ntoa(c->client_addr.sin_addr)
		<< "-" << c->client_addr.sin_port
		<< "\n";
       }
   cout << "\n";
   
   // Now, print the contents of the log on primary.
   cout << "Log Content\n";
   cout << "============\n";
   cmdPacket *tmp;
   for ( unsigned int i = 0; i < log.size(); i++ )
       {
	   char *buf = log[i];
	   tmp = reinterpret_cast<cmdPacket*>(buf);
	   cout << "Log Entry Num:" << i
		<< " Entry Index:" << tmp->logIndex
		<< " File:" << tmp->fileName
		<< " Offset:" << tmp->offset
		<< "\n";	   
       }
   cout << "\n";
   
   // Finally, display the offsets that have been flushed to the
   // backup for all the files in the system
   cout << "Flushed status\n";
   cout << "==============\n";
   map<int, char*>::iterator fIter;
   for ( fIter = flushOffsetTable.begin(); fIter != flushOffsetTable.end(); fIter++)
       {
	   cout << "File-Offset:" << fIter->second
		<< "\n";
       }
   
   delaycb(5, 0, wrap(displayPrimaryStatus));
}


// Send a response to a write from a client.
// The response data will contain randomly selected
// log entries that the client will cache for use
// during disaster recovery
static void sendPRsp(int client_sock)
{
    char *buf = clients[client_sock]->buf; // Buffer contents should still be valid
    cmdPacket *rsp = reinterpret_cast<cmdPacket*>(buf);
    rsp->cmd = PRsp;
    if (clients[client_sock]->flush == true )
	{
	    rsp->flushReq = true;
	    clients[client_sock]->flush = false;
	}

    rsp->logIndex = logCounter++;
    char *newLogEntry = new char[sizeof(cmdPacket)+100];
    memcpy(newLogEntry, buf, sizeof(cmdPacket) + 100);
    log.push_back(newLogEntry);    

	int numBufsToSend = 0;
	char *outBuf = new char[numRspToClient * (sizeof(cmdPacket) + 100) ];
	memset(outBuf, 0, numRspToClient * (sizeof(cmdPacket) + 100));
	// Now, the log entry has been committed. Pick up a
	// random entry from the available log entries and
	// send it back to the client
	char *temp = outBuf;
	while ( numBufsToSend < numRspToClient )
	{
		int numLogEntries = log.size();
		int randomLogEntry = random() % numLogEntries;
		//warn << "Log Entries = " << numLogEntries << " and random:" << randomLogEntry << "\n";
		char *rspBuf = log[randomLogEntry];
		cmdPacket *rspCmd = reinterpret_cast<cmdPacket*>(rspBuf);
		if ( numBufsToSend < numRspToClient-1 )
			rspCmd->moreToFollow = true;			
		memcpy(temp, rspBuf, sizeof(cmdPacket) + 100 );
		// Set the flush request to false, so other clients don't see it
		// again
		rspCmd->flushReq = false;
		temp += (sizeof(cmdPacket) + 100);
		numBufsToSend++;
	}
	    
    int numBytes = write(client_sock, outBuf, numRspToClient * (sizeof(cmdPacket) + 100));

    if ( numBytes < 0 )
	{
	    warn << "Send Rsp: Write error\n";
	    shutdown(client_sock, SHUT_RDWR);
	    fdcb(client_sock, selwrite, 0);
	    close(client_sock);	    
	}
    else if ( numBytes == 0 )
	{
	    warn << "Wrote 0 bytes back to client\n";
	}
    else
	{
	    //warn << "Wrote: " << numBytes << " back to the client\n";
	    clients[client_sock]->readDisabled = true;	    
	    fdcb(client_sock, selwrite, 0);
	}
}

// Read data from a client
static void doClientReads(int client_sock)
{
    char *buf = clients[client_sock]->buf;
    memset(buf, 0, MAX_DATA);
    int n = read(client_sock, buf, 228);
    if ( n < 0 )
	{
	    // Close this connection off
	    warn << "Read error\n";
	    shutdown(client_sock, SHUT_RDWR);
	    fdcb(client_sock, selread, 0);
	    close(client_sock);
	}
    else if ( n == 0 )
	{
	    fdcb(client_sock, selread, 0);	    
	}
    else
	{
	    //warn << "Read:" << n << " bytes from the client\n";
	    cmdPacket *tmp = reinterpret_cast<cmdPacket*>(buf);
	    if ( tmp->cmd == CWrite )
		{
		    //warn << "Received a write from client for file:"
		    // << tmp->fileName << " at offset:"
		    // << tmp->offset << "\n";		    
		    fdcb(client_sock, selread, 0);
		    fdcb(client_sock, selwrite, wrap(sendPRsp,
						     client_sock));
						     
		}
	}
}


// Send back a confirmation message to the client
// that it is now registered with the primary
static void sendRegnConf(int client_sock)
{
    client *regnC = clients[client_sock];
    if ( regnC == (client *)0 )
	{
	    // Close this connection off
	    warn << "Regn. Recv: unable to find client info??\n";
	    shutdown(client_sock, SHUT_RDWR);
	    fdcb(client_sock, selread, 0);
	    close(client_sock);	    
	}    
    cmdPacket *regConf = new cmdPacket;
    memset(regConf, 0, sizeof(cmdPacket));
    regConf->cmd = RegisterDone;
    int n = write(client_sock, regConf, sizeof(cmdPacket));
    if ( n < 0 )
	{
	    // Close this connection off
	    warn << "Regn. Send: write returned error\n";
	    shutdown(client_sock, SHUT_RDWR);
	    fdcb(client_sock, selread, 0);
	    close(client_sock);	    
	    delete regConf;
	    clients.erase(client_sock);	    
	}
    else if ( n == 0 )
	{
	    warn << "Reg. Send: write returned 0\n";
	    delete regConf;
	}
    else
	{
	    regnC->registered = true;
	    delete regConf;
	    // Disable writes on this socket for now
	    fdcb(client_sock, selwrite, 0);
	    // Enable reads on this socket now
	    fdcb(client_sock, selread, wrap(doClientReads, client_sock));
	}
}

// Receive registration information from a client.
// This step will help the primary obtain some useful
// information from a client.
static void receiveRegnInfo(int client_sock)
{
    client *regnC = clients[client_sock];
    if ( regnC == (client *)0 )
	{
	    // Close this connection off
	    warn << "Regn. Recv: unable to find client info??\n";
	    shutdown(client_sock, SHUT_RDWR);
	    fdcb(client_sock, selread, 0);
	    close(client_sock);
	}    
    memset(regnC->buf, 0, sizeof(regnC->buf));
    int n = read(client_sock, regnC->buf, MAX_DATA);
    
    if ( n < 0 )
	{
	    fdcb(client_sock, selread, 0); // Unregister read callback
	    close(client_sock);
	    delete regnC;
	    clients.erase(client_sock);
	    fatal << "Reg. Recv: Read error\n";	   
	}
    else if ( n == 0 )
	{
	    warn << "Reg. Recv: received 0 bytes??\n";	    
	}
    else
	{
	    //warn << "Reg. Recv: received " << n << " bytes from client \n";
	    cmdPacket *cmd = new cmdPacket;
	    memset(cmd, 0, sizeof(cmdPacket));
	    memcpy(cmd, regnC->buf, sizeof(cmdPacket));
	    if ( cmd->cmd != Register )
		{
		    // Close this connection off
		    warn << "Regn. Recv: invalid packet received\n";
		    delete cmd;
		    delete regnC;
		    clients.erase(client_sock);		    
		    shutdown(client_sock, SHUT_RDWR);
		    fdcb(client_sock, selread, 0);
		    close(client_sock);
		}
	    else
		{
		    regnC->clientPort = cmd->localPort;		    
		    //warn << "Regn Recv: Register request recvd. from " <<
		    //     regnC->client_hname << "\n";
		    fstream oFile;
		    oFile.open("/tmp/fuse/clients", fstream::out | fstream::app);
		    char str[50];
		    sprintf(str, "%s\n%d\n", inet_ntoa(clients[client_sock]->client_addr.sin_addr),
			    clients[client_sock]->clientPort);
		    oFile.write(str, strlen(str));
		    oFile.flush();
		    oFile.close();		    
		    

		    delete cmd;
		    // Unset read on this socket - until we send the registration
		    // confirmation
		    fdcb(client_sock, selread, 0);
		    // Now,register a write on this socket, so we can send regn. conf
		    fdcb(client_sock, selwrite, wrap(sendRegnConf, client_sock));
		}
	}
}

// Ah, a client has just sent a connect message to the primary.
// Accept the connection, and await registration from the client.
static void connectFromClient()
{
    int client_sock;
    struct sockaddr_in proxy_addr, client_addr;
    socklen_t addr_len = sizeof(struct sockaddr_in);
    memset(&proxy_addr, 0, sizeof(struct sockaddr_in));	
    if ((client_sock = accept(primarySock, (struct sockaddr *) &proxy_addr, 
			      &addr_len)) < 0)
	{
	    if (errno != EAGAIN) /* nonblocking, retry */
		fprintf(stderr, "accept error %s\n", strerror(errno));
	}
    else
	{
	    make_async(client_sock); 
	    addr_len = sizeof(struct sockaddr_in); /* in/out parameter */
	    if (getpeername(client_sock, (struct sockaddr *) &client_addr, 
			    &addr_len) < 0)
		{
		    fprintf(stderr, "getsockname error %s\n", strerror(errno));
		    exit(1);
		}
	    // Create a new client for this socket
	    client *newC = new client;
	    newC->clientSock = client_sock;
	    strncpy(newC->client_hname, inet_ntoa(client_addr.sin_addr), MAX_ADDR_NAME);
	    //warn << "Connection received:" <<
	    //newC->client_hname << ", Port:" << ntohs(client_addr.sin_port) << "\n";

	    // Insert into our client map
	    clients[client_sock] = newC;
	    memcpy(&clients[client_sock]->client_addr, &client_addr, sizeof(client_addr));
	    
	    // Enable read on the newly created socket - 
	    // we need to receive registration information
	    // from the client
	    
	    fdcb(client_sock, selread, 0); // Unregister old function
	    fdcb(client_sock, selread, wrap(receiveRegnInfo, client_sock)); // Register new function
	}    
}

// Read the contents of a client's cache, and apply 
// the entries to the backup
static void readLogFromClient(int cSock)
{
    char logData[MAX_DATA];
    memset(logData, 0, sizeof(logData));
    int n = read(cSock, logData, MAX_DATA);
    if ( n < 0 )
	{
	    fatal << "Recovery failed while reading client response!\n";
	}
    else if ( n == 0 )
	{
	    fatal << "Recovery failed while reading client response!\n";   
	}
    else
	{
	    //warn << "Read:" << n << " bytes from this client\n";
	    char *readEntry = logData;
	    while ( n > 0 )
		{
		    cmdPacket *tmp = reinterpret_cast<cmdPacket*>(readEntry);
		    warn << "Received Log entry index:" << tmp->logIndex
			 << ", filename:" << tmp->fileName
			 << ", offset:" << tmp->offset
		     << "\n";
		    int fd = open(tmp->fileName, O_WRONLY);
		    if ( fd < 0 )
			{
			    cout << "Unable to open file:" << tmp->fileName
				 << " for writing!\n";
			    readEntry += (sizeof(cmdPacket) + 100 );
			    n -= (sizeof(cmdPacket) + 100 );			    
			    continue;
			}
		    lseek(fd, tmp->offset, SEEK_SET);
		    write(fd, readEntry + sizeof(cmdPacket), 100);
		    int endOff = lseek(fd, 0, SEEK_END);
		    cout << "Current end offset for "
			 << tmp->fileName 
			 <<" this file:" << endOff;
		    recoveredFiles[tmp->fileName] = endOff;
		    close(fd);
		    readEntry += (sizeof(cmdPacket) + 100 );
		    n -= (sizeof(cmdPacket) + 100 );
		}
	    fdcb(cSock, selread, 0);
	} 
}

// Send a query message to the client, asking for the contents
// of its logs.
static void sendLogQueryToClient(int cSock)
{
    cmdPacket comm;
    memset(&comm, 0, sizeof(cmdPacket));
    comm.cmd = LogQuery;
    int n = write(cSock, &comm, sizeof(cmdPacket));
    if ( n < 0 )
	{
	    fatal << "Recovery failed while issuing LogQuery!\n";
	}
    else if ( n == 0 )
	{
	    fatal << "Recovery failed while issuing LogQuery!\n";   
	}
    else
	{
	    fdcb(cSock, selwrite, 0);
	    fdcb(cSock, selread, wrap(readLogFromClient, cSock));
	}
}


// Connect to a client so we could check to see if it holds
// any data in its cache that can be used for recovery.
static void connectToRecoveryClient(int recoverySock, struct sockaddr_in rClient_addr)
{
    if ( connect(recoverySock, (struct sockaddr *) &rClient_addr, 
		 sizeof(struct sockaddr_in)) < 0)
	{
	    if (errno != EINPROGRESS) /* nonblocking, connection stalled */
		{
		    fprintf(stderr, "connect error %s\n", strerror(errno));
		    close(recoverySock);
		    exit(1);
		}
	}
    else
	{
	    cout << "Connected to " << inet_ntoa(rClient_addr.sin_addr) << "\n";
	    fdcb(recoverySock, selwrite, 0);
	    fdcb(recoverySock, selwrite, wrap(sendLogQueryToClient, recoverySock));
	}		    		        
}


// The recovery process
// Read list of clients from backup metadata
// Connect to each client in turn, and read
// log data, if any, from them.
void recoverFromClients()
{
    fstream iFile;
    iFile.open("/tmp/fuse/clients", fstream::in);
    char ipLine[200];
    char portLine[200];
    cout << "\33[H\33[2J";    
    cout << "\t\t\tStarting Recovery Process\n";
    if ( iFile.is_open() )
	{
	    while(!iFile.eof() )
		{
		    memset(ipLine, 0, 200);
		    memset(portLine, 0, 200);
		    iFile.getline(ipLine, 200);
		    if ( strlen(ipLine) < 5 ) 
			break;
		    warn << "Recover from IP:" << ipLine << "\t";		    
		    iFile.getline(portLine, 200);
		    int port = atoi(portLine);
		    warn << "port:" << port << "\n";		    
		    if ( port <= 0 )
			break;
		    struct sockaddr_in rClient_addr;
		    memset(&rClient_addr, 0, sizeof(struct sockaddr_in));
		    inet_aton(ipLine, &(rClient_addr.sin_addr));
		    rClient_addr.sin_port = htons(port);
		    rClient_addr.sin_family = AF_INET;		    

		    int recoverySock;
		    
		    if ((recoverySock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0 )
			{
			    fatal << "Unable to create recovery socket: Exiting!\n";
			    exit(1);
			}

		    struct sockaddr_in recovery_addr;
		    memset(&recovery_addr, 0, sizeof(struct sockaddr_in));
		    recovery_addr.sin_family = AF_INET;
		    recovery_addr.sin_addr.s_addr = INADDR_ANY; /* bind to all local addresses */
		    recovery_addr.sin_port = htons(recovery_port++);
		    if (bind(recoverySock, (struct sockaddr *) &recovery_addr, 
			     sizeof(recovery_addr)) < 0)
			{
			    fprintf(stderr, "bind error %s\n", strerror(errno));
			    exit(1);
			}    
		    make_async(recoverySock);		    
		    // Now, connect to the client with this info
		    recoverySockList.push_back(recoverySock);
		    fdcb(recoverySock, selwrite, wrap(connectToRecoveryClient,
						      recoverySock,
						      rClient_addr));
		}
	}
    amain();
    iFile.close();
}

// Start routine.
// If we are in recovery mode, jump straight to
// recovery code, else start as a primary process.
// The two modes are not connected unfortunately,
// hence, you will first have to run this code in
// non-recovery mode, and then, start it in recovery
// mode to recover from the last instance that it failed.
int main(int argc, char** argv)
{
    struct sockaddr_in primary_addr;
    int local_port = PRIMARY_PORT;

    if ( argc != 3 )
	{
	    fatal << "Usage: primary <opt> <numRsp>\n"
	    	  << "where opt = 0 for no recovery, 1 to recover\n"
	    	  << "numRsp(1-3) is the number of tags sent in a response to clients\n";
	}

    cout << "\33[H\33[2J";
    cout.flush();
    int recover = atoi(argv[1]);
    numRspToClient = atoi(argv[2]);
    if ( numRspToClient < 1 || numRspToClient > 3 )
    {
    	fatal << "<numRsp> must be between 1-3\n";
    }
    
    if ( recover )
	{
	    recoverFromClients();
	}

    if ( !recover )
	{
	    // Truncate data file containing all the client ids
	    truncate("/tmp/fuse/clients", 0);
	    if ((primarySock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0 )
		{
		    fatal << "Unable to create primary socket: Exiting!\n";
		    exit(1);
		}

		// Create a primary socket that will be used to communicate
		// with clients.
	    memset(&primary_addr, 0, sizeof(struct sockaddr_in));
	    primary_addr.sin_family = AF_INET;
	    primary_addr.sin_addr.s_addr = INADDR_ANY; /* bind to all local addresses */
	    primary_addr.sin_port = htons(local_port);
	    if (bind(primarySock, (struct sockaddr *) &primary_addr, 
		     sizeof(primary_addr)) < 0)
		{
		    fprintf(stderr, "bind error %s\n", strerror(errno));
		    exit(1);
		}
    
		make_async(primarySock);
		listen(primarySock, MAX_LISTEN_BACKLOG);
		/* register socket and signal handlers with the async event loop */
		fdcb(primarySock, selread, wrap(connectFromClient));
		sigcb(SIGINT, wrap(exit, 1));
		sigcb(SIGTERM, wrap(exit, 1));

	    delaycb(45, 0, wrap(flushWrites));
	    delaycb(5, 0, wrap(displayPrimaryStatus));
	    amain(); /* async main */
	}
    return 0;
}
