/*
 * Filesystem using Amazon S3
 * Author: Avinash Kulkarni
 * NetID: ak678
 * Original shell program credits: Tudor Marian
 */
#include "s3fs.h"
#include "s3ops.h"
#include "metadata.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>

//#define NDEBUG /* disables assert if macro defined above #include */
#include <assert.h>

struct s3fs_buckets *sb;
superblock *localSB;          // Local copy of SB
inode *localRoot;             // Local copy of root node
fileEnt *localRootData;       // Local copy of root data

struct dirbuf
{
  char *p;
  size_t size;
};

/* inode number 1 belongs to /, fuse assumes this */

/* format filesystem --- superblock, and inode 1 (or /) */
int s3fs_format(char *inode_bucket, char *data_bucket)
{
  // Reset the buckets to initial state
 
#ifdef _DEBUG
  fprintf(stderr, "Formatting the buckets\n");
#endif

 
  s3_clear_bucket(inode_bucket);
  s3_clear_bucket(data_bucket);
  return 1;
}

/* 
 * initialize filesystem --- e.g. read the superblock
 * @param userdata the user data passed to fuse_lowlevel_new()
 */
void s3fs_init(void *userdata, struct fuse_conn_info *conn)
{
  // Allocate space to store pointers to 2 buckets
  sb = (struct s3fs_buckets *) malloc(sizeof(struct s3fs_buckets));
  memcpy(sb, userdata, sizeof(struct s3fs_buckets));

  // Create a super block object in the MD bucket
  localSB = (superblock *)malloc(sizeof(superblock));
  memset(localSB, 0, sizeof(superblock));
  localSB->ino_num_used = 2;
  localSB->numFiles = 1;

  // Write out the superblock object to the MD bucket
  ssize_t wrote = s3_put_object(sb->inode, "0", localSB, sizeof(*localSB));
  if ( wrote <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "Write error to SB\n");
#endif

      return;
    }
  
  // Create a new localRoot inode object, and a localRoot data object
  localRoot = (inode *)malloc(sizeof(inode));
  memset(localRoot, 0, sizeof(localRoot));
  localRoot->stat.st_ino = 1;
  localRoot->stat.st_mode = S_IFDIR | 0777;
  localRoot->stat.atime = localRoot->stat.ctime = localRoot->stat.mtime = time(NULL);
  localRoot->stat.st_size = 4096;
  localRoot->numFiles = 1;

  // Place the localRoot inode into MD bucket
  wrote = s3_put_object(sb->inode, "1", localRoot, sizeof(inode));
  if ( wrote <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "Write error to localRoot inode\n");
#endif

      return;
    }


  // Create a root data object
  localRootData = (fileEnt *)malloc(sizeof(fileEnt));
  memset(localRootData, 0, sizeof(fileEnt));
  memcpy(localRootData->fileName, ".", strlen("."));

  // Place the localRoot inode into MD bucket
  wrote = s3_put_object(sb->data, "1", localRootData, sizeof(fileEnt));
  if ( wrote <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "Write error to localRoot data file\n");
#endif

      return;
    }  
}


// Get stat information about the inode 'ino' passed
// in
static int do_stat(fuse_ino_t ino, struct stat *stbuf)
{
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(stbuf));
  char inoStr[3];
  sprintf(inoStr, "%d", (int)ino);

  // Read in inode from MD bucket for this inode
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      free(temp);
#ifdef _DEBUG
      fprintf(stderr, "Failed lookup on node:%s", inoStr);
#endif

      return -1;
    }

  // Fill in the information from inode
  stbuf->st_mode = temp->stat.st_mode;
  stbuf->st_ino = temp->stat.st_ino;
  stbuf->st_atime = temp->stat.atime;
  stbuf->st_mtime = temp->stat.mtime;
  stbuf->st_ctime = temp->stat.ctime;
  if (S_ISDIR(temp->stat.st_mode))
    stbuf->st_nlink = 2;
  stbuf->st_size = temp->stat.st_size;
#ifdef _DEBUG
  fprintf(stderr, "Size of file:%d is %d\n", (int)ino, temp->stat.st_size);
#endif

  free(temp);
  return 0;
}

// Get file attributes for specified inode
void s3fs_getattr(fuse_req_t req, fuse_ino_t ino,
		  struct fuse_file_info *fi)
{
  struct stat stbuf;
  memset(&stbuf, 0, sizeof(stbuf));

  if (do_stat(ino, &stbuf) == -1)
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "-1 from stat?\n");
#endif

    }
  else
    fuse_reply_attr(req, &stbuf, 1.0);
}

// Check to see if 'name' filename exists in the directory 
// inode pointed to by parent
void s3fs_lookup(fuse_req_t req, fuse_ino_t parent, const char *name)
{
  struct fuse_entry_param e;
  char inoStr[3];

 
#ifdef _DEBUG
  fprintf(stderr, "Lookup request:p:%d, name:%s\n", (int)parent, name);
#endif

 

  // Check to see if the parent inode exists in MD bucket
  sprintf(inoStr, "%d", (int)parent);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      free(temp);
      fuse_reply_err(req, ENOENT);
      return;
    }
      
  // Check if the parent inode is a directory
  if (!S_ISDIR(temp->stat.st_mode))
    {
      free(temp);
      fuse_reply_err(req, ENOTDIR);
      return;
    }

  // Get the number of files in the directory
  int numFiles = temp->numFiles;
 
#ifdef _DEBUG
  fprintf(stderr, "Parent has %d files\n", numFiles);
#endif

 
  free(temp);

  // Get the data block for this inode
  fileEnt *curFileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 1));
  memset(curFileBuf, 0, sizeof(sizeof(fileEnt) * (numFiles + 1)));
  r = s3_read_object(sb->data, inoStr, curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
      return;
    }

  // Loop through the file entries in the data block of
  // of the directory inode and see if this file already
  // exists.
  fileEnt *iter = curFileBuf;
  int found = 0;
  int inoNum = 0;
  int i = 0;
  while ( (i++ < numFiles) && (strcmp(iter->fileName, ".") != 0 ))
    {
 
#ifdef _DEBUG
      fprintf(stderr, "%s\n", iter->fileName);
#endif

 
      if (strcmp(iter->fileName, name) == 0 )
	{
	  found = 1;
	  inoNum = iter->ino_num;
	  break;
	}
      iter++;
    }

  free(curFileBuf);
  if ( found == 0 )
    {
      fuse_reply_err(req, ENOENT);
      return;
    }
  
  memset(&e, 0, sizeof(e));
  e.ino = inoNum;
  e.attr_timeout = 1.0;
  e.entry_timeout = 1.0;
  do_stat(e.ino, &e.attr);
  fuse_reply_entry(req, &e);

}

/* Miklos's original dirbuf_add code, I am not a big fan of realloc */
static void dirbuf_add(fuse_req_t req, struct dirbuf *b, const char *name,
		       fuse_ino_t ino)
{
  struct stat stbuf;
  size_t oldsize = b->size;
  b->size += fuse_add_direntry(req, NULL, 0, name, NULL, 0);
  char *newp = (char *)realloc(b->p, b->size);
  if (!newp)
    {
#ifdef _DEBUG
      fprintf(stderr, "*** fatal error: cannot allocate memory\n");
#endif

      abort();
    }
  b->p = newp;
  memset(&stbuf, 0, sizeof(stbuf));
  stbuf.st_ino = ino;
  fuse_add_direntry(req, b->p + oldsize, b->size - oldsize, name, &stbuf,
		    b->size);
}

#define min(x, y) ((x) < (y) ? (x) : (y))

static int reply_buf_limited(fuse_req_t req, const char *buf, size_t bufsize,
			     off_t off, size_t maxsize)
{
 
#ifdef _DEBUG
  fprintf(stderr, "Off:%d, bufsz:%d\n", (int)off, bufsize);
#endif

 
  if (off < bufsize)
    return fuse_reply_buf(req, buf + off, min(bufsize - off, maxsize));
  else
    return fuse_reply_buf(req, NULL, 0);
}

// Get all the files that are contained in the directory
// pointed to by 'ino'
void s3fs_readdir(fuse_req_t req, fuse_ino_t ino, size_t size,
		  off_t off, struct fuse_file_info *fi)
{
  char inoStr[3];
  sprintf(inoStr, "%d", (int)ino);

 
#ifdef _DEBUG
  fprintf(stderr, "Listing contents of dir ino:%s, size:%d, off:%d\n", inoStr, (int)size, (int)off );
#endif

 

  // Read in the parent inode from MD bucket
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
      return;
    }
      
  if (!S_ISDIR(temp->stat.st_mode))
    {
      fuse_reply_err(req, ENOTDIR);
      return;
    }
  int numFiles = temp->numFiles;
  free(temp);

  // Get the data block for this inode
  fileEnt *curFileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 1));
  memset(curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  r = s3_read_object(sb->data, inoStr, curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
      free(temp);
      free(curFileBuf);
      return;
    }

  
  int byteCount = 0;
  struct dirbuf b;
  if ( off == 0 )
    {
      // If entering this function for the first
      // time for this directory, then add in the '.'
      // and '..' entries to output buffer
      memset(&b, 0, sizeof(b));
      dirbuf_add(req, &b, ".", 1);
      byteCount += strlen(".");
      dirbuf_add(req, &b, "..", 1);
      byteCount += strlen("..");
    }

  fileEnt *iter = curFileBuf;
  int offCount = 0;
  while ( strcmp(iter->fileName, ".") != 0 )
    {
      // Reach the offset that is being asked for
      if ( offCount < off )
	{
	  offCount += strlen(iter->fileName);
	  iter++;
	  continue;
	}
      
      // Once the offset has been reached, start copying
      // file entries to output buffer, but only until
      // 'size' bytes are already in the output buffer
      if ( byteCount + strlen(iter->fileName) > size )
	break;

 
#ifdef _DEBUG
      fprintf(stderr, "Read %s from dir:%s\n", iter->fileName, inoStr);
#endif

 

      dirbuf_add(req, &b, iter->fileName, iter->ino_num);
      byteCount += strlen(iter->fileName);
      iter++;
    }      

  free(curFileBuf);      
  reply_buf_limited(req, b.p, b.size, off, size);

  // Free the buffer only if something was copied in
  if ( byteCount > 0 )
    {
      free(b.p);
    }

}

// Create a new file(inode+data) in the directory pointed to by 
// parent parameter
void s3fs_mknod(fuse_req_t req, fuse_ino_t parent, const char *name,
		mode_t mode, dev_t rdev)
{
  // Get parent inode, and get parent's data
  char inoStr[3];

 
#ifdef _DEBUG
  fprintf(stderr, "MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  // Read in the parent inode from MD bucket
  sprintf(inoStr, "%d", (int)parent);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "1:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      return;
    }
      
  // Check to see if the parent inode is a directory
  if (!S_ISDIR(temp->stat.st_mode))
    {
      fuse_reply_err(req, ENOTDIR);
#ifdef _DEBUG
      fprintf(stderr, "2:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      return;
    }
  int numFiles = temp->numFiles;
 
#ifdef _DEBUG
  fprintf(stderr, "3:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  // Get the data block for this inode - curFileBuf will contain
  // file entries that the current buffer holds
  fileEnt *curFileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 1));
  memset(curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  r = s3_read_object(sb->data, inoStr, curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "4:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(curFileBuf);
      return;
    }

  // Create new buffer to hold the entry for the new file  
  fileEnt *fileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 2));
  memset(fileBuf, 0, (sizeof(fileEnt) * (numFiles + 2)));

  fileEnt *sBuf = curFileBuf;
  fileEnt *dBuf = fileBuf;

  // Generate a new inode number
  int newInoNum = localSB->ino_num_used++;
  dBuf->ino_num = newInoNum;
  strcpy(dBuf->fileName, name);
  dBuf++;
  // Copy over the file entries from old buffer
  // to the new buffer
  while ( strcmp(sBuf->fileName, ".") != 0 )
    {
      memcpy(dBuf, sBuf, sizeof(fileEnt));
      dBuf++;
      sBuf++;
    }
  // Copy over the last entry
  memcpy(dBuf, sBuf, sizeof(fileEnt));
 
#ifdef _DEBUG
  fprintf(stderr, "5:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  free(curFileBuf);
  // Retain the new parent inode structure - we will flush it last

  // Create a new inode for this file
  inode *temp1 = (inode *)malloc(sizeof(inode));
  memset(temp1, 0, sizeof(inode));  
  temp1->stat.st_ino = newInoNum;
  temp1->stat.st_mode = mode;
  temp1->stat.atime = temp1->stat.ctime = temp1->stat.mtime = time(NULL);
  temp1->stat.rdev = rdev;

  //Create a new data node for this inode
  fileEnt *f;
  if ( S_ISDIR(mode))
    {
      f = (fileEnt *)malloc(sizeof(fileEnt));
      memset(f, 0, sizeof(fileEnt));
      f->ino_num = newInoNum;
      strcpy(f->fileName, ".");
      // Flush this file down to data bucket
      sprintf(inoStr, "%d", newInoNum);
      ssize_t w = s3_put_object(sb->data, inoStr, f, sizeof(fileEnt));
      if ( w <= 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "6:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

	  free(f);
	  free(temp);
	  free(temp1);
	  free(fileBuf);
#ifdef _DEBUG
	  fprintf(stderr, "Failed to write data block for new inode:%s\n", inoStr);
#endif

	  return;
	}
      free(f);
    }
  else
    {
      // Write a small 1 byte field to the data block. 
      char testData[] = "1";
      sprintf(inoStr, "%d", newInoNum);
      ssize_t w = s3_put_object(sb->data, inoStr, testData, 1);
      if ( w <= 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "7:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

	  free(temp);
	  free(temp1);
	  free(fileBuf);
#ifdef _DEBUG
	  fprintf(stderr, "Failed to write data block for new inode:%s\n", inoStr);
#endif

	  return;
	}
    }

  // Flush its inode back to inode bucket
  sprintf(inoStr, "%d", newInoNum);
  ssize_t w = s3_put_object(sb->inode, inoStr, temp1, sizeof(inode));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "8:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to write inode block for new inode:%s\n", inoStr);
#endif

      return;
    }
  
  // Now, flush the parent inode block back into the inode bucket
  sprintf(inoStr, "%d", (int)parent);
  temp->numFiles +=1;
  w = s3_put_object(sb->inode, inoStr, temp, sizeof(inode));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "9:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to parent inode for new inode:%s\n", inoStr);
#endif

      return;
    }
  // And, flush the parent data block back into the data bucket
  sprintf(inoStr, "%d", (int)parent);
  w = s3_put_object(sb->data, inoStr, fileBuf, sizeof(fileEnt)*(numFiles + 2));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "9:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to parent inode for new inode:%s\n", inoStr);
#endif

      return;
    }
  
  // Finally, flush the superblock back into the inode bucket
  sprintf(inoStr, "%d", 0);
  w = s3_put_object(sb->inode, inoStr, &localSB, sizeof(superblock));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "10:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to superblock for new inode:%s\n", inoStr);
#endif

      return;
    }

 
#ifdef _DEBUG
  fprintf(stderr, "11:MKNOD: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  free(temp);
  free(temp1);
  free(fileBuf);

  struct fuse_entry_param e;
  memset(&e, 0, sizeof(e));
  e.ino = newInoNum;
  e.attr_timeout = 1.0;
  e.entry_timeout = 1.0;
  do_stat(e.ino, &e.attr);
  fuse_reply_entry(req, &e);

  return;
}

// Create a new file
void s3fs_create(fuse_req_t req, fuse_ino_t parent, const char *name,
			mode_t mode, struct fuse_file_info *fi)
{
  // Check to see if parent directory exists
  struct fuse_entry_param e;
  memset(&e, 0, sizeof(e));
  e.ino = parent;
  e.attr_timeout = 1.0;
  e.entry_timeout = 1.0;
  if ( do_stat(e.ino, &e.attr) == -1 )
    {
      fuse_reply_err(req, ENOENT);
      return;
    }

    // Get parent inode, and get parent's data
  char inoStr[3];

 
#ifdef _DEBUG
  fprintf(stderr, "CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  sprintf(inoStr, "%d", (int)parent);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "1:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      return;
    }
      
  if (!S_ISDIR(temp->stat.st_mode))
    {
      fuse_reply_err(req, ENOTDIR);
#ifdef _DEBUG
      fprintf(stderr, "2:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      return;
    }
  int numFiles = temp->numFiles;
 
#ifdef _DEBUG
  fprintf(stderr, "3:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 


  // Get the data block for this inode
  fileEnt *curFileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 1));
  memset(curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  r = s3_read_object(sb->data, inoStr, curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "4:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(curFileBuf);
      return;
    }

#ifdef _DEBUG
  fprintf(stderr, "First entry of curFileBuf:%s\n", curFileBuf->fileName);
#endif

  // Create new buffer to hold the entry for the new file
  fileEnt *fileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 2));
  memset(fileBuf, 0, (sizeof(fileEnt) * (numFiles + 2)));

  fileEnt *sBuf = curFileBuf;
  fileEnt *dBuf = fileBuf;

  int newInoNum = localSB->ino_num_used++;
  dBuf->ino_num = newInoNum;
  strcpy(dBuf->fileName, name);
  dBuf++;
  while ( strcmp(sBuf->fileName, ".") != 0 )
    {
      memcpy(dBuf, sBuf, sizeof(fileEnt));
      dBuf++;
      sBuf++;
    }
  memcpy(dBuf, sBuf, sizeof(fileEnt));
 
#ifdef _DEBUG
  fprintf(stderr, "5:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  free(curFileBuf);
  // Retain the new parent inode structure - we will flush it last

  // Create a new inode for this file
  inode *temp1 = (inode *)malloc(sizeof(inode));
  memset(temp1, 0, sizeof(inode));  
  temp1->stat.st_ino = newInoNum;
  temp1->stat.atime = temp1->stat.ctime = temp1->stat.mtime = time(NULL);

#ifdef _DEBUG
  fprintf(stderr, "5.5:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

  
  //Create a new data node for this inode
  fileEnt *f;
  if ( S_ISDIR(mode))
    {
      temp->stat.st_mode = S_IFDIR | 0777;
      temp->stat.st_size = 4096;
      f = (fileEnt *)malloc(sizeof(fileEnt));
      memset(f, 0, sizeof(fileEnt));
      f->ino_num = newInoNum;
      strcpy(f->fileName, ".");
      // Flush this file down to data bucket
      sprintf(inoStr, "%d", newInoNum);
      ssize_t w = s3_put_object(sb->data, inoStr, f, sizeof(fileEnt));
      if ( w <= 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "6:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

	  free(f);
	  free(temp);
	  free(temp1);
	  free(fileBuf);
#ifdef _DEBUG
	  fprintf(stderr, "Failed to write data block for new inode:%s\n", inoStr);
#endif

	  return;
	}
      free(f);
    }
  else
    {
      temp1->stat.st_mode = S_IFREG | 0666;
      char testData[] = "1";
      char inoStr1[3];
      sprintf(inoStr1, "%d", newInoNum);
      ssize_t w = s3_put_object(sb->data, inoStr1, testData, 1);
#ifdef _DEBUG
      fprintf(stderr, "6.5:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      if ( w <= 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "7:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

	  free(temp);
	  free(temp1);
	  free(fileBuf);
#ifdef _DEBUG
	  fprintf(stderr, "Failed to write data block for new inode:%s\n", inoStr);
#endif

	  return;
	}
    }
  // Flush its inode back to inode bucket
  char inoStr2[3];
  sprintf(inoStr2, "%d", newInoNum);
  ssize_t w = s3_put_object(sb->inode, inoStr2, temp1, sizeof(inode));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "8:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to write inode block for new inode:%s\n", inoStr);
#endif

      return;
    }
  
  // Now, flush the parent data block back into the bucket
  char inoStr3[3];
  sprintf(inoStr3, "%d", (int)parent);
  temp->numFiles +=1;
  w = s3_put_object(sb->inode, inoStr3, temp, sizeof(inode));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "9:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to parent inode for new inode:%s\n", inoStr);
#endif

      return;
    }
  char inoStr4[3];
  sprintf(inoStr4, "%d", (int)parent);
  w = s3_put_object(sb->data, inoStr4, fileBuf, sizeof(fileEnt)*(numFiles + 2));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "9:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to parent inode for new inode:%s\n", inoStr);
#endif

      return;
    }
  
  // Finally, flush the superblock back into the inode bucket
  char inoStr5[3];
  sprintf(inoStr5, "%d", 0);
  w = s3_put_object(sb->inode, inoStr5, &localSB, sizeof(superblock));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "10:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to superblock for new inode:%s\n", inoStr);
#endif

      return;
    }
 
#ifdef _DEBUG
  fprintf(stderr, "11:CREATE: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  free(temp);
  free(temp1);
  free(fileBuf);

  memset(&e, 0, sizeof(e));
  e.ino = newInoNum;
  e.attr_timeout = 1.0;
  e.entry_timeout = 1.0;
  if ( do_stat(e.ino, &e.attr) == -1 )
    {
      fuse_reply_err(req, ENOENT);
      return;
    }  

  fuse_reply_create(req, &e, fi);
}

void s3fs_mkdir(fuse_req_t req, fuse_ino_t parent, const char *name,
		mode_t mode)
{
      // Get parent inode, and get parent's data
  char inoStr[3];
 
#ifdef _DEBUG
  fprintf(stderr, "MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 


  mode = S_IFDIR | 0777;

  sprintf(inoStr, "%d", (int)parent);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "1:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      return;
    }
      
  if (!S_ISDIR(temp->stat.st_mode))
    {
      fuse_reply_err(req, ENOTDIR);
#ifdef _DEBUG
      fprintf(stderr, "2:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      return;
    }
  int numFiles = temp->numFiles;
 
#ifdef _DEBUG
  fprintf(stderr, "3:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  // Get the data block for this inode
  fileEnt *curFileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 1));
  memset(curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  r = s3_read_object(sb->data, inoStr, curFileBuf, 0, (sizeof(fileEnt) * (numFiles + 1)));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "4:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(curFileBuf);
      return;
    }
 
#ifdef _DEBUG
  fprintf(stderr, "3.5:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 
  // Create new buffer to hold the entry for the new file
  fileEnt *fileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles + 2));
  memset(fileBuf, 0, sizeof(fileEnt) * (numFiles + 2));

  fileEnt *sBuf = curFileBuf;
  fileEnt *dBuf = fileBuf;

  int newInoNum = localSB->ino_num_used++;
  dBuf->ino_num = newInoNum;
  strcpy(dBuf->fileName, name);
  dBuf++;
  int i = 0;
  while ( (i++ < numFiles) && (strcmp(sBuf->fileName, ".") != 0 ))
    {
      memcpy(dBuf, sBuf, sizeof(fileEnt));
      dBuf++;
      sBuf++;
    }
  memcpy(dBuf, sBuf, sizeof(fileEnt));
 
#ifdef _DEBUG
  fprintf(stderr, "5:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  free(curFileBuf);
 
#ifdef _DEBUG
  fprintf(stderr, "5.25:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 
  // Retain the new parent inode structure - we will flush it last

  // Create a new inode for this file
  inode *temp1 = (inode *)malloc(sizeof(inode));
  memset(temp1, 0, sizeof(inode));  
  temp1->stat.st_ino = newInoNum;
  temp1->stat.st_mode = mode;
  temp1->stat.atime = temp1->stat.ctime = temp1->stat.mtime = time(NULL);
  temp1->stat.st_size = 4096;

  //Create a new data node for this inode
  fileEnt *f;
  if ( S_ISDIR(mode))
    {
 
#ifdef _DEBUG
      fprintf(stderr, "5.5:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

      f = (fileEnt *)malloc(sizeof(fileEnt) * 2);      
      fileEnt *fTemp = f;
      memset(f, 0, sizeof(fileEnt)*2);
      f->ino_num = parent;
      strcpy(f->fileName, "..");
      f++;
      f->ino_num = newInoNum;
      strcpy(f->fileName, ".");
      // Flush this file down to data bucket
      sprintf(inoStr, "%d", newInoNum);
      ssize_t w = s3_put_object(sb->data, inoStr, f, sizeof(fileEnt) * 2);
      if ( w <= 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "6:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

	  free(f);
	  free(temp);
	  free(temp1);
	  free(fileBuf);
#ifdef _DEBUG
	  fprintf(stderr, "Failed to write data block for new inode:%s\n", inoStr);
#endif

	  return;
	}
      free(fTemp);
    }

   
#ifdef _DEBUG
  fprintf(stderr, "7:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  // Flush its inode back to inode bucket
  sprintf(inoStr, "%d", newInoNum);
  ssize_t w = s3_put_object(sb->inode, inoStr, temp1, sizeof(inode));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "8:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to write inode block for new inode:%s\n", inoStr);
#endif

      return;
    }
  
  // Now, flush the parent data block back into the bucket
  sprintf(inoStr, "%d", (int)parent);
  temp->numFiles +=1;
  w = s3_put_object(sb->inode, inoStr, temp, sizeof(inode));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "9:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to parent inode for new inode:%s\n", inoStr);
#endif

      return;
    }
  sprintf(inoStr, "%d", (int)parent);
  w = s3_put_object(sb->data, inoStr, fileBuf, sizeof(fileEnt)*(numFiles + 2));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "9:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to parent inode for new inode:%s\n", inoStr);
#endif

      return;
    }
  
  // Finally, flush the superblock back into the inode bucket
  sprintf(inoStr, "%d", 0);
  w = s3_put_object(sb->inode, inoStr, &localSB, sizeof(superblock));
  if ( w <= 0 )
    {
#ifdef _DEBUG
      fprintf(stderr, "10:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

      free(temp);
      free(temp1);
      free(fileBuf);
#ifdef _DEBUG
      fprintf(stderr, "Failed to superblock for new inode:%s\n", inoStr);
#endif

      return;
    }

#ifdef _DEBUG
  fprintf(stderr, "11:MKDIR: p:%d, name:%s, mode:%x\n", (int)parent, name, mode);
#endif

 

  free(temp);
  free(temp1);
  free(fileBuf);

  struct fuse_entry_param e;
  memset(&e, 0, sizeof(e));
  e.ino = newInoNum;
  e.attr_timeout = 1.0;
  e.entry_timeout = 1.0;
  if ( do_stat(e.ino, &e.attr) == -1 )
    {
      fuse_reply_err(req, ENOENT);
      return;
    }    
  fuse_reply_entry(req, &e);
}

void s3fs_rmdir(fuse_req_t req, fuse_ino_t parent, const char *name)
{
  // Get the parent's inode from metadata bucket
  char inoStr[3];
  sprintf(inoStr, "%d", (int)parent);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory does not exist!\n");
#endif

      free(temp);
      return;
    }  

  if ( !S_ISDIR(temp->stat.st_mode))
    {
      fuse_reply_err(req, ENOTDIR);
#ifdef _DEBUG
      fprintf(stderr, "Parent is not a directory!\n");
#endif

      free(temp);
      return;
    }      

  // Get the data block for this directory file
  int numFiles = temp->numFiles;
  // No point in checking for 'name' if parent directory
  // empty
  if ( numFiles == 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent is not a directory!\n");
#endif

      free(temp);
      return;
    }
#ifdef _DEBUG
  fprintf(stderr, "Reading in inode\n");
#endif

  fileEnt *f = (fileEnt*)malloc(sizeof(fileEnt)*(numFiles + 1));
  r = s3_read_object(sb->data, 
		     inoStr, 
		     f,
		     0, 
		     (sizeof(fileEnt) * (numFiles + 1)));
#ifdef _DEBUG
  fprintf(stderr, "Read in inode\n");
#endif

  if ( r <= 0 )
    {
      // Declare mis-trial
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory data does not exist!\n");
#endif

      free(temp);
      free(f);
      return;
    }

  // Now that we have the data for this directory, check to see
  // if this file exists in the directory, and remove it if found
  fileEnt *iter = f;
  int i = 0;
  int found = 0;  
  while ( i++ < numFiles )
    {
#ifdef _DEBUG
      fprintf(stderr, "i:%d, numFiles:%d\n", i, numFiles);
#endif

      if ( strcmp(iter->fileName, ".") == 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "Check loop logic!!!\n");
#endif

	}
      if ( strcmp(iter->fileName, name) == 0 )
	{
	  found = 1;
	  break;
	}
      iter++;
    }

  if (!found)
    {
      // Declare unsuccessful search
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory data does not exist!\n");
#endif

      free(temp);
      free(f);
      return;
    }

  // Now, read in the inode block of the found entry, and check
  // to see if it is NOT a directory, 
  int inoNum = iter->ino_num;
  char inoStr1[3];
  sprintf(inoStr1, "%d", inoNum);
  inode *temp1 =  (inode *)malloc(sizeof(inode));
  r = s3_read_object(sb->inode, 
		     inoStr1,
		     temp1,
		     0,
		     sizeof(inode));
  if ( r <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Child directory data does not exist!\n");
#endif

      free(temp);
      free(f);
      free(temp1);
      return;
    }
  
  // Check to see if the inode belongs to a directory
  if ( !S_ISDIR(temp1->stat.st_mode))
    {
      // Declare failure
      fuse_reply_err(req, ENOTDIR);
#ifdef _DEBUG
      fprintf(stderr, "File is NOT a directory\n");
#endif

      free(temp);
      free(f);
      free(temp1);
      return;
    }

  // Check to see if the child directory is non-empty
  if ( temp1->numFiles > 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOTEMPTY);
#ifdef _DEBUG
      fprintf(stderr, "Directory is not empty\n");
#endif

      free(temp);
      free(f);
      free(temp1);
      return;
    }
  
#ifdef _DEBUG
  fprintf(stderr, "Found an entry!\n");
#endif

  // Ok, so there is a entry for the directory to be deleted,
  // and it is empty. Delete the entry, and blow away the data
  // block for this entry. Also, remove its entry from parent
  // directory's fileEnt.
  // First remove the entry from parent's directory. This
  // way, even if the rest of the operations fail, the directory
  // entry is no longer in the filesystem. Its definately a key leak,
  // but life sucks right?
  
  // iter already points to the correct entry. Copy over all
  // entries other than iter to a new structure
  fileEnt *src = f;
  fileEnt *newEntries = (fileEnt*)malloc(sizeof(fileEnt)*numFiles); // One entry less
  fileEnt *dst = newEntries;
  i = 0;
  while ( i++ < numFiles )
    {      
      if ( strcmp(src->fileName, ".") == 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "Copy logic is working well\n");
#endif

	}
      if ( src == iter )
	{
	  src++;
	  continue;
	}
      memcpy(dst, src, sizeof(fileEnt));
      dst++;
      src++;      
    }
  free(f); // Don't need f anymore.
  temp->numFiles -= 1;
  
  // Flush parent inode and data block down to the buckets
  ssize_t w = s3_put_object(sb->inode, 
			    inoStr,
			    temp,
			    sizeof(inode));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to parent inode\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      return;      
    }
  
  w = s3_put_object(sb->data,
			    inoStr,
			    newEntries,
			    (sizeof(fileEnt)*numFiles));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to parent data block\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      return;      
    }

  // Now. delete the actual inode and data blocks of the target directory
  char inoStr2[3];
  sprintf(inoStr2, "%d", temp1->stat.st_ino);
  int status = s3_rm_object(sb->inode, inoStr2);
  if ( !status )
    {
#ifdef _DEBUG
      fprintf(stderr, "Unable to delete child inode\n");
#endif

      // But can't return from here. No harm in
      // trying to delete child data block
    }
  status = s3_rm_object(sb->data, inoStr2);
  if ( !status )
    {
#ifdef _DEBUG
      fprintf(stderr, "Unable to delete child data block\n");
#endif

    }

  // Free all the used buffers, and return - FUSE apparently
  // doesn't need success status here?
  free(temp);
  free(temp1);
  free(newEntries);
  fuse_reply_err(req, 0);
  return;
}


void s3fs_unlink(fuse_req_t req, fuse_ino_t parent, const char *name)
{
  // Get the parent's inode from metadata bucket
  char inoStr[3];
  sprintf(inoStr, "%d", (int)parent);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory does not exist!\n");
#endif

      free(temp);
      return;
    }  

  if ( !S_ISDIR(temp->stat.st_mode))
    {
      fuse_reply_err(req, ENOTDIR);
#ifdef _DEBUG
      fprintf(stderr, "Parent is not a directory!\n");
#endif

      free(temp);
      return;
    }      

  // Get the data block for this directory file
  int numFiles = temp->numFiles;
  // No point in checking for 'name' if parent directory
  // empty
  if ( numFiles == 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent is not a directory!\n");
#endif

      free(temp);
      return;
    }
#ifdef _DEBUG
  fprintf(stderr, "Reading in inode\n");
#endif

  fileEnt *f = (fileEnt*)malloc(sizeof(fileEnt)*(numFiles + 1));
  r = s3_read_object(sb->data, 
		     inoStr, 
		     f,
		     0, 
		     (sizeof(fileEnt) * (numFiles + 1)));
#ifdef _DEBUG
  fprintf(stderr, "Read in inode\n");
#endif

  if ( r <= 0 )
    {
      // Declare mis-trial
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory data does not exist!\n");
#endif

      free(temp);
      free(f);
      return;
    }

  // Now that we have the data for this directory, check to see
  // if this file exists in the directory, and remove it if found
  fileEnt *iter = f;
  int i = 0;
  int found = 0;  
  while ( i++ < numFiles )
    {
#ifdef _DEBUG
      fprintf(stderr, "i:%d, numFiles:%d\n", i, numFiles);
#endif

        if ( strcmp(iter->fileName, ".") == 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "Check loop logic!!!\n");
#endif

	}
      if ( strcmp(iter->fileName, name) == 0 )
	{
	  found = 1;
	  break;
	}
      iter++;
    }

  if (!found)
    {
      // Declare unsuccessful search
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory data does not exist!\n");
#endif

      free(temp);
      free(f);
      return;
    }  

  // Now, read in the inode block of the found entry, and check
  // to see if it is a directory, 
  int inoNum = iter->ino_num;
  char inoStr1[3];
  sprintf(inoStr1, "%d", inoNum);
  inode *temp1 =  (inode *)malloc(sizeof(inode));
  r = s3_read_object(sb->inode, 
		     inoStr1,
		     temp1,
		     0,
		     sizeof(inode));
  if ( r <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Child INODE does not exist!\n");
#endif

      free(temp);
      free(f);
      free(temp1);
      return;
    }
  // Check to see if the inode belongs to a directory
  if ( S_ISDIR(temp1->stat.st_mode))
    {
      // Declare failure
      fuse_reply_err(req, EISDIR);
#ifdef _DEBUG
      fprintf(stderr, "File is a directory\n");
#endif

      free(temp);
      free(f);
      free(temp1);
      return;
    }

#ifdef _DEBUG
  fprintf(stderr, "Found an entry!\n");
#endif

  // Ok, so there is a entry for the file to be deleted,
  // and it is empty. Delete the entry, and blow away the data
  // block for this entry. Also, remove its entry from parent
  // directory's fileEnt.
  // First remove the entry from parent's directory. This
  // way, even if the rest of the operations fail, the directory
  // entry is no longer in the filesystem. Its definately a key leak,
  // but life sucks right?
  
  // iter already points to the correct entry. Copy over all
  // entries other than iter to a new structure
  fileEnt *src = f;
  fileEnt *newEntries = (fileEnt*)malloc(sizeof(fileEnt)*numFiles); // One entry less
  fileEnt *dst = newEntries;
  i = 0;
  while ( i++ < numFiles+1 )
    {      
      if ( strcmp(src->fileName, ".") == 0 )
	{
#ifdef _DEBUG
	  fprintf(stderr, "Copy logic is working well\n");
#endif

	}
      if ( src == iter )
	{
	  src++;
	  continue;
	}
      memcpy(dst, src, sizeof(fileEnt));
      dst++;
      src++;      
    }
  free(f); // Don't need f anymore.
  temp->numFiles -= 1;
  
  // Flush parent inode and data block down to the buckets
  ssize_t w = s3_put_object(sb->inode, 
			    inoStr,
			    temp,
			    sizeof(inode));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to parent inode\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      return;      
    }
  
  w = s3_put_object(sb->data,
			    inoStr,
			    newEntries,
			    (sizeof(fileEnt)*numFiles));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to parent data block\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      return;      
    }

  // Now. delete the actual inode and data blocks of the target directory
  char inoStr2[3];
  sprintf(inoStr2, "%d", temp1->stat.st_ino);
  int status = s3_rm_object(sb->inode, inoStr2);
  if ( !status )
    {
#ifdef _DEBUG
      fprintf(stderr, "Unable to delete child inode\n");
#endif

      // But can't return from here. No harm in
      // trying to delete child data block
    }
  status = s3_rm_object(sb->data, inoStr2);
  if ( !status )
    {
#ifdef _DEBUG
      fprintf(stderr, "Unable to delete child data block\n");
#endif

    }

  // Free all the used buffers, and return 
  free(temp);
  free(temp1);
  free(newEntries);
  fuse_reply_err(req, 0);
  return;
}


void s3fs_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
		int to_set, struct fuse_file_info *fi)
{
  ssize_t w;
  // Get the inode from metadata bucket
  char inoStr[3];
  sprintf(inoStr, "%d", (int)ino);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Inode does not exist!\n");
#endif

      free(temp);
      return;
    }  

  int change = 0; // Assume no changes will be made

  if ( to_set & FUSE_SET_ATTR_ATIME )
    {
#ifdef _DEBUG
      fprintf(stderr, "atime:%x\n", (int)attr->st_atime);
#endif

      temp->stat.atime = attr->st_atime;
      change = 1;
    }
  if ( to_set & FUSE_SET_ATTR_MTIME )
    {
#ifdef _DEBUG
      fprintf(stderr, "mtime:%x\n", (int)attr->st_mtime);
#endif

      temp->stat.mtime = attr->st_mtime;
      change = 1;
    }  

  if ( to_set & FUSE_SET_ATTR_SIZE )
    {
#ifdef _DEBUG
      fprintf(stderr, "Old Size: %d, New Size:%d\n", temp->stat.st_size, (int)attr->st_size);      
#endif

      if ( attr->st_size == 0 )
	{
	  // File will be truncated. Just write a single
	  // byte to the key value in S3 data bucket.
	  char tempBuf[] = "1";
	  w = s3_put_object(sb->data, 
			    inoStr,
			    tempBuf,
			    sizeof(tempBuf));
	  if ( w <= 0 )
	    {
	      // Declare failure
	      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
	      fprintf(stderr, "Unable to write to data bucket - truncate case\n");
#endif

	      free(temp);
	      return;      
	    }
	  temp->stat.st_size=0;
	  change = 1;
	}
      else if ( temp->stat.st_size > attr->st_size )
	{	  
	  // Read in the file, and write back only attr->st_size
	  // bytes
	  char *tempBuf = (char *)malloc(sizeof(char) * temp->stat.st_size);
	  memset(tempBuf, 0,sizeof(char) * temp->stat.st_size);
	  r = s3_read_object(sb->data, inoStr, tempBuf, 0, (sizeof(char) * temp->stat.st_size));
	  if ( r <= 0 )
	    {
	      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
	      fprintf(stderr, "Failed to reduce size - unable to read data!\n");
#endif

	      free(temp);
	      free(tempBuf);
	      return;
	    }
	  // Now, writeback only attr->st_size bytes
	  w = s3_put_object(sb->data, 
			    inoStr,
			    tempBuf,
			    attr->st_size);
	  if ( w <= 0 )
	    {
	      // Declare failure
	      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
	      fprintf(stderr, "Unable to write to data bucket - truncate case\n");
#endif

	      free(temp);
	      free(tempBuf);
	      return;      
	    }
	  free(tempBuf);	  
	  temp->stat.st_size =  attr->st_size;
	  change = 1;
	}
      else if ( temp->stat.st_size < attr->st_size )
	{
	  // Create a new buffer of attr->st_size bytes
	  // and fill in the initial temp->stat.st_size bytes
	  // with S3 contents. Pad up the rest of bytes with 0
	  // and flush the buffer to data bucket
	  char *tempBuf = (char *)malloc(sizeof(char) * attr->st_size);
	  memset(tempBuf, 0,sizeof(char) * attr->st_size);
	  ssize_t r = s3_read_object(sb->data, inoStr, tempBuf, 0, (sizeof(char) * temp->stat.st_size));
	  if ( r <= 0 )
	    {
	      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
	      fprintf(stderr, "Failed to increase size - unable to read data!\n");
#endif

	      free(temp);
	      free(tempBuf);
	      return;
	    }
	  // Rest of the bytes would be set to zeroes anyway
	  // Now, writeback all attr->st_size bytes
	  ssize_t w = s3_put_object(sb->data, 
				    inoStr,
				    tempBuf,
				    attr->st_size);
	  if ( w <= 0 )
	    {
	      // Declare failure
	      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
	      fprintf(stderr, "Unable to write to data bucket - truncate case\n");
#endif

	      free(temp);
	      free(tempBuf);
	      return;      
	    }
	  free(tempBuf);	  
	  temp->stat.st_size =  attr->st_size;
	  change = 1;
	}      
    }
  if ( change == 1 )
    {
      // Something has been changed for this file. 
      // Lets flush the inode back to inode bucket
      w = s3_put_object(sb->inode, 
			inoStr,
			temp,
			sizeof(inode));
      if ( w <= 0 )
	{
	  // Declare failure
	  fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
	  fprintf(stderr, "Unable to write to inode bucket\n");
#endif

	  free(temp);
	  return;      
	}
    }
  else
    {
#ifdef _DEBUG
      fprintf(stderr, "No change request was made????\n");
#endif

    }
  free(temp);
  struct fuse_entry_param e;
  memset(&e, 0, sizeof(e));
  e.ino = ino;
  e.attr_timeout = 1.0;
  e.entry_timeout = 1.0;
  if ( do_stat(e.ino, &e.attr) == -1 )
    {
      fuse_reply_err(req, ENOENT);
      return;
    }    
  fuse_reply_attr(req, &e.attr, 1.0);
}

void s3fs_write(fuse_req_t req, fuse_ino_t ino, const char *buf,
		       size_t size, off_t off, struct fuse_file_info *fi)
{
  ssize_t w;
  // Get the inode from metadata bucket
  char inoStr[3];
  sprintf(inoStr, "%d", (int)ino);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Inode does not exist!\n");
#endif

      free(temp);
      return;
    }    

#ifdef _DEBUG
  fprintf(stderr, "fi->flags: %x, off:%d, size:%d\n", fi->flags, (int)off, (int)size);
#endif


  if ( off > temp->stat.st_size )
    {
#ifdef _DEBUG
      fprintf(stderr, "Warning - off:%d is past file size:%d\n", (int)off, temp->stat.st_size);
#endif
    }
  if ( (off+size) > temp->stat.st_size )
    {
#ifdef _DEBUG
      fprintf(stderr, "Warning - off:%d + size:%d is past file size:%d\n",(int)off, size, temp->stat.st_size);
#endif
    }


  int padding = 0;
  if ( (off + size) > temp->stat.st_size )
    {
      padding = (off + size ) - temp->stat.st_size;
#ifdef _DEBUG
      fprintf(stderr, "Padding:%d bytes are non zero!\n", padding);
#endif

    }
 
  // Allocate a new buffer of fileSize
  char *newBuf = 0 ;
  newBuf = (char *)malloc(sizeof(char) * (temp->stat.st_size + padding));
  if ( !newBuf)
    {
#ifdef _DEBUG
      fprintf(stderr, "Malloc failed in write - Out of memory??\n");
#endif

      free(temp);      
      fuse_reply_err(req, EIO);
      return;
    }
  memset(newBuf, 0, (temp->stat.st_size + padding));
  // Read in the data contents into the new buffer - but
  // only if atleast one write has been done to the file ever
  if ( temp->stat.st_size > 0 )
    {
      r = s3_read_object(sb->data, inoStr, newBuf, 0, temp->stat.st_size);
      if ( r <= 0 )
	{
	  fuse_reply_err(req, EIO);
#ifdef _DEBUG
	  fprintf(stderr, "Data does not exist!\n");
#endif

	  free(temp);
	  free(newBuf);
	  return;
	}
    }    

  // Now, overwrite the region of file buffer that overlaps the
  // write buffer that was passed in
  memcpy(newBuf+off, buf, size);
  
  // Now, flush the data back into data bucket
  w = s3_put_object(sb->data, 
		    inoStr,
		    newBuf,
		    (temp->stat.st_size + padding));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, EIO);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to data  bucket\n");
#endif

      free(temp);
      free(newBuf);
      return;      
    }

  temp->stat.st_size += padding;
  temp->stat.atime = temp->stat.ctime = temp->stat.mtime = time(NULL);
  // Flush the inode back into inode bucket
  w = s3_put_object(sb->inode, 
		    inoStr,
		    temp,
		    sizeof(inode));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, EIO);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to data  bucket\n");
#endif

      free(temp);
      free(newBuf);
      return;      
    }
  
  free(temp);
  fuse_reply_write(req, size);
  free(newBuf);
  return;
}

void s3fs_read(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off,
		struct fuse_file_info *fi)
{
#ifdef _DEBUG
  fprintf(stderr, "In read function - size:%d, off:%d!\n", size, (int)off);
#endif

  char inoStr[3];
  sprintf(inoStr, "%d", (int)ino);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Inode does not exist!\n");
#endif

      free(temp);
      return;
    }    
 
  int readSize = size;

  // Read past EOF - or file never written to
  if ( (off >= temp->stat.st_size) || (temp->stat.st_size == 0))
    {
      free(temp);      
      char *tempBuf1 =  (char *)malloc(sizeof(char));
      tempBuf1[0] = '\0';
      fuse_reply_buf(req, tempBuf1, 0);
      free(tempBuf1);
      return;
    }
  
  // Read starts within file, but ends at EOF. Truncate read
  if (( off + size ) > temp->stat.st_size)
    {
      readSize = temp->stat.st_size - off;
#ifdef _DEBUG
      fprintf(stderr, "\nReadsize:%d\n", readSize);
#endif

    }


  char *tempBuf =  (char *)malloc(sizeof(char) * (size));
  memset(tempBuf, 0, size);  
  r = s3_read_object(sb->data, inoStr, tempBuf, off, readSize);
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Error in data read!\n");
#endif

      free(temp);
      free(tempBuf);
      return;
    }    

  // Flush the inode back into inode bucket
  temp->stat.atime = time(NULL);
  ssize_t w = s3_put_object(sb->inode,
			    inoStr,
			    temp,
			    sizeof(inode));
  if ( w <= 0 )
    {
      // Warn about failure to flush inode
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to inode bucket\n");
#endif
    }


  free(temp);
  fuse_reply_buf(req, tempBuf, readSize);
  free(tempBuf);
  return;
}


void s3fs_rename(fuse_req_t req, fuse_ino_t parent, const char *name,
		fuse_ino_t newparent, const char *newname)
{
  // Read in parent inode. Check to see if valid dir
#ifdef _DEBUG
  fprintf(stderr, "In rename function - p:%d, n:%s, np:%d, nn:%s\n", (int)parent, name, (int)newparent, newname);
#endif

  char inoStr[3];
  sprintf(inoStr, "%d", (int)parent);
  inode *temp = (inode *)malloc(sizeof(inode));
  memset(temp, 0, sizeof(inode));
  ssize_t r = s3_read_object(sb->inode, inoStr, temp, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Inode does not exist!\n");
#endif

      free(temp);
      return;
    }    

  int numFiles = temp->numFiles;
  if ( numFiles == 0 )
    {
      // Can't rename anything if no entries in the directory
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Inode does not exist!\n");
#endif

      free(temp);
      return;
    }
  
  // Read in parent data. Check for presence of name
#ifdef _DEBUG
  fprintf(stderr, "Reading in data for parent inode\n");
#endif

  fileEnt *f = (fileEnt*)malloc(sizeof(fileEnt)*(numFiles + 1));
  r = s3_read_object(sb->data, 
		     inoStr, 
		     f,
		     0, 
		     (sizeof(fileEnt) * (numFiles + 1)));
  if ( r <= 0 )
    {
      // Declare mis-trial
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory data does not exist!\n");
#endif

      free(temp);
      free(f);
      return;
    }

  // Now that we have the data for this directory, check to see
  // if this file exists in the directory, and remove it if found
  fileEnt *iter = f;
  int i = 0;
  int found = 0;  
  while ( i++ < numFiles )
    {
      if ( strcmp(iter->fileName, name) == 0 )
	{
	  found = 1;
	  break;
	}
      iter++;
    }

  if (!found)
    {
      // Declare unsuccessful search
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Parent directory data does not contain entry for %s\n", name);
#endif

      free(temp);
      free(f);
      return;
    }    

  // Remove name from parent data
  fileEnt *src = f;
  fileEnt *newEntries = (fileEnt*)malloc(sizeof(fileEnt)*numFiles); // One entry less
  fileEnt *dst = newEntries;
  i = 0;
  int renameInoNum = 0;
  while ( i++ < numFiles+1 )
    {      
      if ( strcmp(src->fileName, ".") == 0 )
#ifdef _DEBUG
	fprintf(stderr, "Copy logic looks good\n");
#endif

      if ( src == iter )
	{
	  renameInoNum = src->ino_num;
	  src++;
	  continue;
	}
      memcpy(dst, src, sizeof(fileEnt));
      dst++;
      src++;      
    }
  free(f); // Don't need f anymore.
  temp->numFiles -= 1;
  // Now, flush old parent inode
  ssize_t w = s3_put_object(sb->inode, 
		    inoStr,
		    temp,
		    sizeof(inode));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to inode bucket\n");
#endif

      free(temp);
      free(newEntries);
      return;      
    } 
  // Last, old parent data
  w = s3_put_object(sb->data, 
		    inoStr,
		    newEntries,
		    (sizeof(fileEnt)*numFiles));

  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to inode bucket\n");
#endif

      free(temp);
      free(newEntries);
      return;      
    } 
  
  // Read in newparent inode. Check to see if valid dir
  ////////////////////////////////
  char inoStr1[3];
  sprintf(inoStr1, "%d", (int)newparent);
  inode *temp1 = (inode *)malloc(sizeof(inode));
  memset(temp1, 0, sizeof(inode));
  r = s3_read_object(sb->inode, inoStr1, temp1, 0, sizeof(inode));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "1:rename\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      return;
    }
      
  if (!S_ISDIR(temp->stat.st_mode))
    {
      fuse_reply_err(req, ENOTDIR);
#ifdef _DEBUG
      fprintf(stderr, "2:rename\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      return;
    }
  int numFiles1 = temp1->numFiles;

  // Get the data block for this inode
  fileEnt *curFileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles1 + 1));
  memset(curFileBuf, 0, (sizeof(fileEnt) * (numFiles1 + 1)));
  r = s3_read_object(sb->data, inoStr1, curFileBuf, 0, (sizeof(fileEnt) * (numFiles1 + 1)));
  if ( r <= 0 )
    {
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "3:rename\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      free(curFileBuf);
      return;
    }

  // Create new buffer to hold the entry for the new file
  fileEnt *fileBuf = (fileEnt *)malloc(sizeof(fileEnt) * (numFiles1 + 2));
  memset(fileBuf, 0, sizeof(fileEnt) * (numFiles1 + 2));

  fileEnt *sBuf = curFileBuf;
  fileEnt *dBuf = fileBuf;

  dBuf->ino_num = renameInoNum;
  strcpy(dBuf->fileName, newname);
  dBuf++;
  i = 0;
  while ( (i++ < numFiles1) && (strcmp(sBuf->fileName, ".") != 0 ))
    {
      memcpy(dBuf, sBuf, sizeof(fileEnt));
      dBuf++;
      sBuf++;
    }
  memcpy(dBuf, sBuf, sizeof(fileEnt));  
  temp1->numFiles += 1;
  ////////////////////////////////  

  // First, flush new parent inode
  w = s3_put_object(sb->inode, 
		    inoStr1,
		    temp1,
		    sizeof(inode));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to inode bucket\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      free(curFileBuf);
      free(fileBuf);
      return;      
    } 
  // Next, new parent data
  w = s3_put_object(sb->data, 
		    inoStr1,
		    fileBuf,
		    (sizeof(fileEnt) * (numFiles1 + 2)));
  if ( w <= 0 )
    {
      // Declare failure
      fuse_reply_err(req, ENOENT);
#ifdef _DEBUG
      fprintf(stderr, "Unable to write to inode bucket\n");
#endif

      free(temp);
      free(temp1);
      free(newEntries);
      free(curFileBuf);
      free(fileBuf);
      return;      
    } 


  fuse_reply_err(req, 0);
}
