/*
 * @author Tudor Marian <tudorm@cs.cornell.edu>
 */

#ifndef _S3_FS_H_
#define _S3_FS_H_

#ifdef __cplusplus
extern "C"
{
#endif

/* before including fuse_lowlevel.h */
#define FUSE_USE_VERSION 26
/* this header is pretty much the only documentation available */
#include <fuse_lowlevel.h>
#include <string.h>

struct s3fs_buckets
{
	char *inode;
	char *data;
};

/* format your s3fs */
int s3fs_format(char *inode_bucket, char *data_bucket);

/* called by fuse_lowlevel_new with userdata param */
void s3fs_init(void *userdata, struct fuse_conn_info *conn);

/* man 2 path_resolution */
void s3fs_lookup(fuse_req_t req, fuse_ino_t parent, const char *name);

/* man 2 stat ( also check http://en.wikipedia.org/wiki/Stat_(Unix) ) */
void s3fs_getattr(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi);

/* man 2 stat, man 2 chmod */
void s3fs_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
		int to_set, struct fuse_file_info *fi);

/* man 2 mknod */
void s3fs_mknod(fuse_req_t req, fuse_ino_t parent, const char *name,
		mode_t mode, dev_t rdev);

/* man 2 mkdir */
void s3fs_mkdir(fuse_req_t req, fuse_ino_t parent, const char *name,
		mode_t mode);

/* man 2 unlink */
void s3fs_unlink(fuse_req_t req, fuse_ino_t parent, const char *name);

/* man 2 rmdir */
void s3fs_rmdir(fuse_req_t req, fuse_ino_t parent, const char *name);

/* man 2 rename --- annoyingly intricate
 * in principle, like unlink/rmdir + mknod/mkdir modulo corner cases Fuse 
 * takes care of the fact that newparent may not be a subdirectory of parent.
 */
void s3fs_rename(fuse_req_t req, fuse_ino_t parent, const char *name,
		fuse_ino_t newparent, const char *newname);

/* man 2 open -- do not use if filesystem is stateless */
void s3fs_open(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi);

/* man 2 close -- do not use if filesystem is stateless */
void s3fs_release(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi);

/* man 2 read */
void s3fs_read(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off,
		struct fuse_file_info *fi);

/* man 2 write, also check ``man 2 lseek'' for out of bounds write */
void s3fs_write(fuse_req_t req, fuse_ino_t ino, const char *buf,
		       size_t size, off_t off, struct fuse_file_info *fi);

/* man 2 readdir */
void s3fs_readdir(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off,
		struct fuse_file_info *fi);

/* man 2 creat --- just like mknod + open, or mknod for a stateless fs */
void s3fs_create(fuse_req_t req, fuse_ino_t parent, const char *name,
			mode_t mode, struct fuse_file_info *fi);

static inline void s3fs_ops_init(struct fuse_lowlevel_ops *ops)
{
	/* do not remove this unless all functions are plugged in the struct */
	memset(ops, 0, sizeof(struct fuse_lowlevel_ops));
	
	/* Fill in all newly implemented operations */
	ops->init = s3fs_init;
	ops->lookup = s3fs_lookup;
	ops->getattr = s3fs_getattr;
	ops->readdir = s3fs_readdir;
	ops->mknod = s3fs_mknod;
	ops->create = s3fs_create;
	ops->mkdir = s3fs_mkdir;
	ops->rmdir = s3fs_rmdir;
	ops->unlink = s3fs_unlink;
	ops->setattr = s3fs_setattr;
	ops->write = s3fs_write;
	ops->read = s3fs_read;
	ops->rename = s3fs_rename;
}

#ifdef __cplusplus
}
#endif

#endif //_S3_FS_H_

